@@ -1,5 +1,5 @@
 /*
-  Copyright (C) 1999-2001 Ricardo Ueda Karpischek
+  Copyright (C) 1999-2002 Ricardo Ueda Karpischek
 
   This is free software; you can redistribute it and/or modify
   it under the terms of the version 2 of the GNU General Public
@@ -48,11 +48,26 @@ int lksz=0,toplk=-1;
 
 /*
 
+Type of action to perform, when the user trains an already
+classified symbol. The user selects one from the following
+codes:
+
+1 - both the pattern and symbol equal the old transliteration
+2 - the symbol transliteration is the new, the pattern is the old
+3 - both the pattern and symbol equal the new transliteration
+
+*/
+int action_type = 1;
+char *r_tr_old,*r_tr_new;
+int r_tr_patt,r_tr_symb;
+
+/*
+
 Collect data to process the submission of transliteration and
 symbol properties.
 
 */
-void review_tr(adesc *a)
+int review_tr(adesc *a)
 {
     /*
         The id of the pattern of the class.
@@ -60,11 +75,11 @@ void review_tr(adesc *a)
     int b;
 
     /*
-        "the class was untransliterated", "the symbol is the pattern",
-        "the symbol is classified", "the symbol was bad" and "the
-        symbol is bad" statuses
+        "the class is untransliterated", "the symbol is classified",
+        "the symbol is not the pattern of the class", "the symbol was bad"
+        and "the symbol is bad" statuses
     */
-    int u,q,c,wb,ib;
+    int u,c,np,wb,ib;
 
     /* the old transliteration and the old flags */
     char *old_tr;
@@ -74,7 +89,10 @@ void review_tr(adesc *a)
     char *new_tr;
 
     /* other */
-    int i,old_mc,old_cdfc;
+    int i,old_mc,old_cdfc,toolarge;
+
+    /* step */
+    static int st = 1;
 
     /* (devel)
 
@@ -88,12 +106,22 @@ void review_tr(adesc *a)
         a transliteration may change the transliteration of the
         current symbol, and also the transliterations of all
         symbols on its class. Depending on the properties,
-        other actions may be performed as well. In order to not
-        inquire the user about what to do, a "protocol"
-        determines what to do in each case. This protocol tries
-        to emulate the "obvious" thing to do in each case,
-        however advanced users are invited to be aware about
-        it:
+        other actions may be performed as well.
+
+    */
+
+    /*
+
+        Decision table (partial)
+        ------------------------
+
+        symbol is classified       0 1 1 1 1
+        tr. differs from class     - 0 1 1 1
+        user input                 - - 1 2 3
+        isnotbad                   X * 0 X 0
+
+        train class                - 0 0 0 1
+        add patt                   X 0 0 X 0
 
     */
 
@@ -112,6 +140,46 @@ void review_tr(adesc *a)
 
     */
 
+    /*
+        Decide about asking the user or not.
+    */
+    if (st == 1) {
+        int bm,p;
+        char *bm_tr;
+
+        /*
+            If the symbol is classified and the submitted
+            transliteration differ from the class transliteration,
+            then ask the user about what to do.
+        */
+        if ((!batch_mode) &&
+            (a->tr != NULL) &&
+            ((bm=mc[a->mc].bm) >= 0) &&
+            ((p=id2idx(bm)) >= 0) &&
+            ((bm_tr=pattern[p].tr) != NULL) &&
+            (strcmp(a->tr,bm_tr) != 0)) {
+
+            setview(PATTERN_ACTION);
+            r_tr_old = bm_tr;
+            r_tr_new = a->tr;
+            r_tr_patt = p;
+            r_tr_symb = a->mc;
+            enter_wait("ENTER or OK to proceed, ESC to cancel",0,2);
+            st = 2;
+            dw[PATTERN_ACTION].rg = 1;
+            return(1);
+        }
+
+        /* otherwise use the default action */
+        action_type = 1;
+    }
+
+    /* must go back to PAGE */
+    if (!batch_mode)
+        setview(PAGE);
+
+    st = 1;
+
     /*
         Save the current values. Maybe some service requires
         curr_mc and cdfc pointing to the current symbol/pattern,
@@ -126,31 +194,37 @@ void review_tr(adesc *a)
     */
     curr_mc = a->mc;
 
+    /* too large? */
+    toolarge = (((mc[curr_mc].r - mc[curr_mc].l + 1) > FS) ||
+                ((mc[curr_mc].b - mc[curr_mc].t + 1) > FS));
+
     /*
-        Check if the class of curr_mc is untransliterated.
+        Check if the symbol is classified and also if the its
+        class (if any) is untransliterated.
     */
     cdfc = id2idx(b=mc[curr_mc].bm);
-    if ((0 <= cdfc) && (cdfc <= topp))
+    if ((0 <= cdfc) && (cdfc <= topp)) {
         u = (pattern[cdfc].tr == NULL);
-    else
+        c = 1;
+    }
+    else {
         u = 0;
+        c = 0;
+    }
 
-    /*
-        Check if the pattern of the class of curr_mc is
-        curr_mc itself.
-    */
-    if (cdfc >= 0) {
-        q = (pattern[cdfc].e == curr_mc) &&
-            (strcmp(pattern[cdfc].d,pagename) == 0);
+    /* check if the symbol is not the pattern of the class */
+    if (c) {
+        unsigned char bm[FS*FS];
+
+        pixel_mlist(a->mc);
+        cb2bm(bm);
+        np = (memcmp(bm,pattern[cdfc].b,FS*FS) != 0);
     }
     else
-        q = 0;
-
-    /* Check if the symbol is classified */
-    c = (mc[curr_mc].bm >= 0);
+        np = 1;
 
     /*
-        Preserve the old transliteration of curr_mc (if any)
+        Remember the old transliteration of curr_mc (if any)
         and the old flags.
     */
     if ((mc[curr_mc].tr == NULL) ||
@@ -174,7 +248,12 @@ void review_tr(adesc *a)
         0. Remove the current revision vote (if any).
 
     */
-    rmvotes(REVISION,curr_mc,-1,0);
+    {
+        int obm;
+        obm = mc[a->mc].bm;
+        rmvotes(REVISION,curr_mc,-1,0);
+        mc[a->mc].bm = obm;
+    }
 
     /* (devel)
 
@@ -188,17 +267,19 @@ void review_tr(adesc *a)
     new_tr = (mc[curr_mc].tr)->t;
 
     /* (devel)
-        2. If the symbol is unclassified and it is not bad,
-        then add its bitmap as a pattern.
+
+        2. If the symbol is unclassified or if the transliteration
+        differ from the class transliteration then add its bitmap
+        as a pattern, unless it's bad or too large.
     */
-    if ((!c) && (!ib)) {
+    if (((!c) || (action_type==2)) && (!ib) && (!toolarge)) {
 
         /*
             This will add the current symbol as a
             pattern.
         */
         update_pattern(curr_mc,new_tr,curr_act,a->a,a->pt,-1,a->f);
-
+	
         /*
             This will make the symbol its own best match.
         */
@@ -207,25 +288,26 @@ void review_tr(adesc *a)
 
     /* (devel)
 
-        3. If the symbol is classified and it was not bad and it's
-        still not bad, propagate the submitted transliteration
-        to the entire class.
+        3. If the symbol is classified and the class is untransliterated
+        or if the user asked to change the class transliteration, then
+        propagate the submitted transliteration to the entire class.
     */
-    else if ((c) && (!wb) && (!ib)) {
+    else if (((c && u) || (action_type==3)) && (!ib) && (!toolarge)) {
 
         /*
             This will change the transliteration of
             the pattern of the class.
         */
         update_pattern(-1,new_tr,curr_act,a->a,a->pt,-1,a->f);
-
+	
         /*
             This will resubmit the SHAPE vote of all
             symbols on the class except curr_mc.
         */
         for (i=0; i<=tops; ++i) {
-            if ((i!=curr_mc) && (mc[i].bm == b))
+            if ((i!=curr_mc) && (mc[i].bm == b)) {
                 add_tr(i,a->tr,SHAPE,curr_act,mc[i].bq,a->a,a->f);
+            }
         }
     }
 
@@ -235,8 +317,7 @@ void review_tr(adesc *a)
         the class, and it was not bad, but it became bad now,
         then remove it from its class, and remove its SHAPE vote.
     */
-    else if ((c) && (!q) && (!wb) && (ib)) {
-
+    else if ((c) && (np) && (!wb) && (ib)) {
         mc[curr_mc].bm = -1;
         rmvotes(SHAPE,curr_mc,-1,0);
     }
@@ -244,25 +325,74 @@ void review_tr(adesc *a)
     /* (devel)
 
         5. If the symbol is the pattern of the class, and it
-        was not bad but it became bad now, then remove it from
+        was not bad, but it became bad now, then remove it from
         the patterns.
     */
+    /*
     else if ((q) && (!wb) && (ib)) {
 
         rm_pattern(cdfc);
     }
+    */
 
     /*
         6. In all other cases generate a debug message.
     */
     else {
-        db("unexpected statuses at review_tr: c=%d,q=%d,wb=%d,ib=%d",
-            c,q,wb,ib);
+        db("unexpected statuses at review_tr: c=%d,wb=%d,ib=%d",c,wb,ib);
     }
 
     /* restore the old values */
     curr_mc = old_mc;
     cdfc = old_cdfc;
+
+    return(0);
+}
+
+/*
+
+Submit pattern transliteration.
+
+*/
+void review_patt(adesc *a)
+{
+    /* other */
+    int i,old_cdfc;
+
+    /* (devel)
+
+        The function review_patt
+        ------------------------
+
+    */
+    old_cdfc = cdfc;
+
+    /*
+        Use the symbol information (if any) as pattern ID.
+    */
+    if (a->mc >= 0)
+        cdfc = id2idx(a->mc);
+    else
+        cdfc = -1;
+
+    /*
+        Change the pattern transliteration.
+    */
+    update_pattern(-1,a->tr,curr_act,a->a,a->pt,-1,a->f);
+
+    /*
+        This will resubmit the SHAPE vote of all
+        symbols on the class except curr_mc.
+    */
+    if (a->mc >= 0) {
+        for (i=0; i<=tops; ++i) {
+            if (mc[i].bm == a->mc)
+                add_tr(i,a->tr,SHAPE,curr_act,mc[i].bq,a->a,a->f);
+        }
+    }
+
+    /* restore the old values */
+    cdfc = old_cdfc;
 }
 
 /*
@@ -276,7 +406,7 @@ void review_merge(adesc *a)
 
     int ws,t;
     int i,j,n,m,*wc,*af;
-    int f;
+    /*int f;*/
 
     k = a->a1;
     l = a->a2;
@@ -326,20 +456,24 @@ void review_merge(adesc *a)
         bookfont and has more than one closure, then that entry
         is marked as a fragment.
     */
+    /*
     if ((mc[l].ncl > 1) && ((f=s2p(l)) >= 0)) {
         C_SET(pattern[f].f,F_FRAG);
     }
     if ((mc[k].ncl > 1) && ((f=s2p(k)) >= 0)) {
         C_SET(pattern[f].f,F_FRAG);
     }
+    */
 
     /*
         if the new symbol belongs to the bookfont, make it a
         complete pattern.
     */
+    /*
     if ((f=s2p(k)) >= 0) {
         C_UNSET(pattern[f].f,F_FRAG);
     }
+    */
 
     /* the merge result becomes the current symbol */
     curr_mc = t;
@@ -405,7 +539,7 @@ Process submission of symbol disassemble.
 */
 void review_dis(adesc *a)
 {
-    int i,*p,f,k;
+    int i,*p/*,f*/,k;
 
     k = a->mc;
 
@@ -423,16 +557,20 @@ void review_dis(adesc *a)
         p = alloca((1+mc[k].ncl) * sizeof(int));
 
         /* the unitary pattern (if any) is unmarked as fragment */
+        /*
         f = s2p(mc[k].cl[i]);
         if (f >= 0) {
             pattern[f].f |= F_FRAG;
             pattern[f].f ^= F_FRAG;
         }
+        */
     }
 
     /* The symbol pattern (if any) is marked as fragment */
+    /*
     f = s2p(k);
     pattern[f].f |= F_FRAG;
+    */
 
     /*
         Must remove k from ps array and include each
@@ -470,31 +608,39 @@ won't be added to the log of acts. Otherwise, will read the XML
 revision record currently in the "text" buffer.
 
 */
-void review(adesc *a)
+int review(int reset,adesc *b)
 {
+    static adesc *a;
+    int r = 0;
 
-    /*
-
-        Read the XML message and use it to initialize the
-        fields of act[topa].
-
-    */
-    if (a == NULL) {
+    if (reset) {
 
         /*
-            Increase topa by 1, extract the fields from the XML
-            message message currently on the buffer text, and use
-            them to initialize act[topa].
-        */
-        recover_acts(NULL);
 
-        a = act + topa;
-        curr_act = topa;
+            Read the XML message and use it to initialize the
+            fields of act[topa].
+
+        */
+        if (b == NULL) {
+
+            /*
+                Increase topa by 1, extract the fields from the XML
+                message message currently on the buffer text, and use
+                them to initialize act[topa].
+            */
+            recover_acts(NULL);
+
+            a = act + topa;
+            curr_act = topa;
+        }
+        else {
+            a = b;
+        }
     }
 
     /* transliteration handler */
     if (a->t == REV_TR)
-        review_tr(a);
+        r = review_tr(a);
 
     /* merge handler */
     else if (a->t == REV_MERGE)
@@ -512,7 +658,14 @@ void review(adesc *a)
     else if (a->t == REV_DIS)
         review_dis(a);
 
-    hist_changed = 1;
+    /* pattern handler */
+    if (a->t == REV_PATT)
+        review_patt(a);
+
+    if (r == 0)
+        hist_changed = 1;
+
+    return(r);
 }
 
 /*
@@ -554,18 +707,19 @@ void reviewer_data(adesc *a)
 
 /*
 
-Entry point for transliterations submitted from keystrokes on the PAGE
-window. The transliteration properties (alphabet, italic and bold
-flags, etc) are read from the interface buttons, and the full data is
-used to generate a XML message containing a revision act informing
-the current symbol, the submitted transliteration and the default
-reviewer info.
+Entry point for transliterations submitted from keystrokes on the
+PAGE or PATTERN windows. The transliteration properties
+(alphabet, italic and bold flags, etc) are read from the
+interface buttons, and the full data is used to generate a XML
+message containing a revision act informing the current symbol,
+the submitted transliteration and the default reviewer info.
 
 */
 void gen_tr(char *tr)
 {
     adesc *a;
     char *ttr;
+    int s;
 
     /* alloc more space for acts */
     if (++topa >= actsz) {
@@ -573,19 +727,25 @@ void gen_tr(char *tr)
     }
     a = act + topa;
 
+    /* symbol or pattern? */
+    s = (to_rev == REV_TR);
+
     /* preamble */
-    a->t  = REV_TR;
-    a->d  = pagename;
+    a->t = s ? REV_TR : REV_PATT;
+    a->d = pagename;
     a->f = 0;
 
-    /* symbol */
-    a->mc = curr_mc;
+    /* symbol or pattern */
+    a->mc = s ? curr_mc : ((cdfc<0) ? -1 : pattern[cdfc].id);
 
     /* properties */
     a->a  = inv_balpha[(int)(button[balpha])];
-    a->f  = ((button[bitalic]) ? F_ITALIC : 0);
+    a->f = ((button[bbad]) ? F_BAD : 0);
+    /*
+    a->f |= ((button[bitalic]) ? F_ITALIC : 0);
     a->f |= ((button[bbold]) ? F_BOLD : 0);
-    a->f |= ((button[bbad]) ? F_BAD : 0);
+    */
+
     /*
     a->pt = (*cm_o_ftro == ' ') ? button[btype] : 0;
     */
@@ -795,8 +955,8 @@ void reset()
             printf("going to read page %s\n",a->d);
 
             /* ugly! */
-            if (load_page(c,1)) {
-                while(load_page(c,0));
+            if (load_page(c,1,0)) {
+                while(load_page(c,0,0));
             }
         }
 
@@ -804,7 +964,8 @@ void reset()
         if (act[i].t == REV_TR) {
             printf("will process act %d\n",i);
             curr_act = i;
-            review(act+i);
+            review(1,act+i);
+            while (review(0,act+i));
         }
 
     }
@@ -986,7 +1147,7 @@ void summarize(int m)
         mc[m].tc = DUBIOUS;
 
     /* compute the alignment of the first transliteration */
-    if ((mc[m].tr != NULL) && (strlen(mc[m].tr->t) == 1)) {
+    if ((mc[m].tr != NULL) /* && (strlen(mc[m].tr->t) >= 1) Obsolete: alignment for multichar transliteration managed. (GL)*/) {
 
         mc[m].va = (mc[m].tr==NULL) ? -1 : tr_align(mc[m].tr->t);
     }
@@ -1041,7 +1202,7 @@ void rmvotes(int o,int k,int a,int nd)
                     /*
                        If the vote origin is SHAPE or REVISION, must clear the
                        best match field. In the first case, because the bm field
-                       stores the patterns from which this vote (being removed)
+                       stores the pattern from which this vote (being removed)
                        was deduced. In the second, because we're just visiting
                        the symbol from which the pattern was derived (the bm
                        field of such symbols is used to store the derived pattern).
@@ -1051,10 +1212,12 @@ void rmvotes(int o,int k,int a,int nd)
 
                     /*
                        If the vote origin is SHAPE, must clear the F_SS
-                       status. OOPS.. not sure abut (!nd).
+                       status and reset lfa.
                     */
-                    if ((!nd) && (v->orig == SHAPE))
+                    if ((!nd) && (v->orig == SHAPE)) {
                         C_UNSET(mc[k].f,F_SS);
+                        mc[k].lfa = -1;
+                    }
 
 		    /* removing vote from symbol k */
                     if (lv == NULL)
@@ -1161,8 +1324,10 @@ parameters to_tr and to_arg. Additional data (italic flag,
 propagability, etc) is taken from the buttons.
 
 */
-void from_gui(void)
+int from_gui(void)
 {
+    int r=1;
+
     /* prepare to submit transliteration */
     if (to_rev == REV_TR) {
 
@@ -1187,6 +1352,34 @@ void from_gui(void)
         next_symb();
     }
 
+    /* prepare to submit pattern transliteration */
+    else if (to_rev == REV_PATT) {
+
+        /*
+            Generates a revision record for the current symbol
+            using the transliteration to_tr, the statuses of the
+            buttons and the nopropag flag.
+        */
+        gen_tr(to_tr);
+
+        /* reset nopropag */
+        /*
+        if (nopropag) {
+            nopropag = 0;
+            mb[0] = 0;
+            redraw_inp = 1;
+        }
+        */
+
+        /*
+            try to move to the next pattern.
+        */
+        if (cdfc < 0)
+            cdfc = topp;
+        else
+            right();
+    }
+
     /* prepare to merge fragment to symbol */
     else if (to_rev == REV_MERGE) {
         int k,l;
@@ -1206,11 +1399,12 @@ void from_gui(void)
 
         /* ignore symbols without at least two components */
         if (mc[k].ncl <= 1)
-            return;
+            r = 0;
 
-        gen_dis(curr_mc);
-
-        curr_mc = mc[k].cl[0];
+        else {
+            gen_dis(curr_mc);
+            curr_mc = mc[k].cl[0];
+        }
     }
 
     /* prepare to link symbols */
@@ -1225,14 +1419,10 @@ void from_gui(void)
 
     /* no operation at all */
     else {
-        return;
+        r = 0;
     }
 
-    /* execute operation */
-    review(NULL);
-
-    /* reset to_rev */
-    to_rev = 0;
+    return(r);
 }
 
 /*
